/*******************************************************************************
 *	Mystique ramdac driver
 *******************************************************************************
 *
   Copyright (C) 1995-1997 The Linux-GGI Project
 *
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *******************************************************************************
 */

#include <kgi/maintainers.h>
#define	MAINTAINER	Matthias_Grimrath
#define	DRIVER_NAME	"Mystique DAC driver"
#define	DRIVER_REV	"$Revision: 1.8 $"

#include <kgi/kgi.h>
#include <kgi/module.h>
#include <kgi/kgi_commands.h>

#include <IBM/vga_dac.h>
#include <Matrox/mga_common_regs.h>
/* Driver specific things are included by mga_common_*.h */
#include <Matrox/mga1064_dac.h>

#include <Matrox/mga_common.h>

static struct kgi_ramdac ramdac = 
{
	"Matrox Graphics Inc",		/* Manufacturer		*/
	"Mystique",		        /* Model		*/
	GM_TEXT16 | GM_8BIT,	        /* Supported modes	*/
	FL_DAC_8BIT,		 	/* Flags		*/
	{ 0xFFFF, 0xFFFF },		/* Maximum resolution	*/
	NULL,				/* Private data hook	*/
	DAC_8BIT_PB,  			/* Device type		*/
	{       0,         0 },         /* Normal mode		*/
	{       0,         0 },		/* 2xclk mode		*/
	{
		{	  0,  55000000 },	/* KGIGT_TEXT16	*/
		{	  0,         0 },	/* KGIGT_TEXT32	*/
		{         0,         0 },	/* KGIGT_1BIT	*/
		{	  0,         0 },	/* KGIGT_4BIT	*/
		{         0,         0 },	/* KGIGT_8BIT	*/	
		{	  0,         0 },	/* KGIGT_15BIT	*/
		{ 	  0,         0 },	/* KGIGT_16BIT	*/
		{         0,         0 },	/* KGIGT_24BIT	*/
		{	  0,         0 }	/* KGIGT_32BIT	*/
	}
};

/*
** Mystique DAC support
*/
void EDAC_out8(struct kgi_display *dpy, kgiu8 val, kgiu8 reg)
{
	GET_PRIVATE;
	ASSERT(reg < 0x50);

	mem_out8(reg, MGA_RAMDAC);
	mem_out8(val, X_DATAREG);
	DEBUG3("DAC mem %.2x <- %.2x", reg, val);
}

kgiu8 EDAC_in8(struct kgi_display *dpy, kgiu8 reg)
{
	GET_PRIVATE;
	register kgiu8 val;

	ASSERT(reg < DAC_MAX_REGS);
	mem_out8(reg, MGA_RAMDAC);
	val = mem_in8(X_DATAREG);
	DEBUG3("DAC mem %.2x -> %.2x", val, reg);
	return val;
}

#define	CLUT_MAGIC	257	/* 2^8 + 2^1 */
#define	CLUT_SHIFT	8

void dac_set_clut(struct kgi_display *dpy, kgi_clut *clut,
		  kgi_uint first, kgi_uint cnt)
{
	GET_PRIVATE;
	kgiu16 *data = (kgiu16 *) (clut->data + first);
	ASSERT(cnt);
	cnt *= 3;
	mem_out8(first, PALWTADD);

	while (cnt--)
	        mem_out8(*data++ >> CLUT_SHIFT, PALDATA);
}

void dac_get_clut(struct kgi_display *dpy,
		  kgi_clut *clut, kgi_uint first, kgi_uint cnt)
{
	GET_PRIVATE;
	kgiu16 *data = (kgiu16 *) (clut->data + first);
	cnt *= 3;
	mem_out8(first, PALRDADD);

	while (cnt--) 
	        *data++ = mem_in8(PALDATA) * CLUT_MAGIC;
}


/* GGI interface */

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	GET_PRIVATE;
	kgi_uint speed;

	NOTICE(DRIVER_NAME " rev. " DRIVER_REV);
	
	/* Set DAC speed dependent on revision */
	if (private->cp.rev < 2) { /* Max 135Mhz */
		speed = 135000000;
	} else if (private->cp.rev < 3) {
		speed = 170000000;
	} else {
		speed = 220000000;
	}

	MGA_CP.chipinfo.dclk.max =
	  ramdac.normal.max =
	  ramdac.clock[KGIGT_8BIT].max =
	  ramdac.clock[KGIGT_15BIT].max =
	  ramdac.clock[KGIGT_16BIT].max =
	  ramdac.clock[KGIGT_24BIT].max =
	  ramdac.clock[KGIGT_32BIT].max = speed;
	
	/* Set DAC to 8 bit, disable feature connector */
	EDAC_out8(dpy, EDAC_in8(dpy, XMISCCTRL) | (MFC_SEL_DIS | VGA_8_DAC),
		  XMISCCTRL);
	
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");

	/* Set DAC to 6 bit, VGA connector output */
	EDAC_out8(dpy, EDAC_in8(dpy, XMISCCTRL) & ~(MFC_SEL_MASK | VGA_8_DAC),
		  XMISCCTRL);
}

int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd *cmd)
{
	kgi_uint newclock = mode->dclk;
  
	switch (mode->request.graphtype) {

	case KGIGT_TEXT16:
	case KGIGT_8BIT:
	case KGIGT_15BIT:
	case KGIGT_16BIT:
	case KGIGT_24BIT:
	case KGIGT_32BIT:
		break;
  
	default:
  		DEBUG2("couldn't handle graphtype %i", 
		       mode->request.graphtype);
		return -E(RAMDAC,MODE_CHECK);
	}

	switch (*cmd) {

	case CMD_PROPOSE:
		mode->lclk.mul = mode->clk.mul = 1;
		mode->lclk.div = mode->clk.div = 1;
		return EOK;

	case CMD_LOWER:

		if ((newclock < ramdac.clock[mode->request.graphtype].min) ||
			(newclock < ramdac.normal.min)) {

			DEBUG2("DCLK too low (%i Hz).", newclock);
			return -E(RAMDAC,CLK_LOW_LIMIT);
		}

		if (ramdac.clock[mode->request.graphtype].max < newclock) {

			newclock = ramdac.clock[mode->request.graphtype].max;
		}
		mode->dclk = newclock;
		return EOK;

	case CMD_RAISE:
		if ((newclock > ramdac.clock[mode->request.graphtype].max) ||
			(newclock > ramdac.normal.max)) {

			DEBUG2("DCLK too high (%i Hz).", newclock);
			return -E(RAMDAC,CLK_HIGH_LIMIT);
		}

		if (newclock < ramdac.clock[mode->request.graphtype].min) {

			newclock = ramdac.clock[mode->request.graphtype].min;
		}

		if (newclock < ramdac.normal.min) {

			newclock = ramdac.normal.min;
		}

		mode->dclk = newclock;
		return EOK;

     case CMD_CHECK:
		if ((mode->lclk.mul != 1) || (mode->lclk.div != 1) ||
		    (mode->clk.mul != 1) || (mode->clk.div != 1)) {
			DEBUG2("clk/lck mul/div bogus");
			return -E(RAMDAC,MODE_CHECK);
		}

		if ((mode->dclk < ramdac.clock[mode->request.graphtype].min) ||
		    (mode->dclk < ramdac.normal.min)) {
			DEBUG2("dclk min");
			return -E(RAMDAC,CLK_LOW_LIMIT);
		}

		if ((mode->dclk > ramdac.clock[mode->request.graphtype].max) ||
		    (mode->dclk > ramdac.normal.max)) {
			DEBUG2("dclk max");			
			return -E(RAMDAC,CLK_HIGH_LIMIT);
		}

		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(RAMDAC,BOGUS);
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	IS_CRITICAL;
	int i,j;
	GET_PRIVATE;

        ENTER_CRITICAL;
	dpy->set_clut = NULL;
        if (mode->magnify.x == 2)
                EDAC_out8(dpy, HZOOM_2X, XZOOMCTRL);
        else
                EDAC_out8(dpy, HZOOM_1X, XZOOMCTRL);

	mem_out8(0xFF, PIXRDMSK);
	switch (mode->request.graphtype) {

	case KGIGT_8BIT:
		EDAC_out8(dpy, XMUL_GT8_PAL, XMULCTRL);
		/* Fall thro */
	case KGIGT_TEXT16:
	case KGIGT_1BIT:
	case KGIGT_4BIT:
		dpy->set_clut = dac_set_clut;
		break;

	case KGIGT_15BIT:
		EDAC_out8(dpy, XMUL_GT15OVL_PAL, XMULCTRL);
		for (j=0; j!=256; j+=128) {
			mem_out8(j, MGA_RAMDAC);
			for (i=0; i!=32; i++) {
				mem_out8(i*8424/1024, PALDATA);
				mem_out8(i*8424/1024, PALDATA);
				mem_out8(i*8424/1024, PALDATA);
			} /* Ignore overlay bit */
		} /* Set default gamma/color correction */
		break;
		
	case KGIGT_16BIT:
		EDAC_out8(dpy, XMUL_GT16_PAL, XMULCTRL);
		mem_out8(0, MGA_RAMDAC);
		for (i=0; i!=255; i++) {
			if (i < 32)
				mem_out8(i*8424/1024, PALDATA);
			else
			        mem_out8(255, PALDATA);
			if (i < 64)
			        mem_out8(i*4145/1024, PALDATA);
			else
			        mem_out8(255, PALDATA);
			if (i < 32)
				mem_out8(i*8424/1024, PALDATA);
			else
			        mem_out8(255, PALDATA);
		} /* Set default gamma/color correction */
		break;

	case KGIGT_24BIT:
		EDAC_out8(dpy, XMUL_GT24_PAL, XMULCTRL);
		goto common_24_32_dacsetup;

	case KGIGT_32BIT:
		EDAC_out8(dpy, XMUL_GT32_PAL, XMULCTRL);
		common_24_32_dacsetup:
		mem_out8(0, MGA_RAMDAC);
		for (i=0; i!=256; i++) {
			mem_out8(i, PALDATA);
			mem_out8(i, PALDATA);
			mem_out8(i, PALDATA);
		} /* Set default gamma/color correction */
		break;
		
	default:
		INTERNAL_ERROR;
	}
        LEAVE_CRITICAL;
}

int kgim_ramdac_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
	/* This may set 
	 * -  Hardware cursor
	 * -  CLUT (or gamma correction table)
	 */
	kgi_clut clut;

	switch (cmd) { 

	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data = (kgi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy, &clut, 0, 256);
		return 0;

	case RAMDAC_GETCLUT:
		clut.size = 256;
		clut.data = (kgi_color *)(graph->io_buf);
		graph->io_res = graph->io_buf;
		dac_get_clut(graph->dev.dpy, &clut, 0, 256);
		return 0;

	default:
		return -E(RAMDAC, NOSUP_ALWAYS_CANT);  
                       /* The ramdac can't do this */
	}
}
